#include "pebblisp.h"
#include "tokens.h"
#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#ifndef STANDALONE
#define printf(...) copySlice(NULL, NULL)
#endif

void copySlice(char * dest, struct Slice *src)
{
    if(!dest || !src)
        return;
    strncpy(dest, src->text, src->length);
    dest[(int)src->length] = '\0';
}

void debugSlice(struct Slice *s)
{
    printf("Debug Slice\n    text:'");
    for(int i = 0; i < s->length; i++) {
        printf("%c", s->text[i]);
    }
    printf("'\n");
    printf("  length: %d\n", s->length);
}

Object fetchFromEnvironment(const char *name, struct Environment *env)
{
    printf("Fetching '%s' from env\n", name);
    int i = 0;
    const char *next = env->strings[i];
    while(next != NULL) {
        if(strcmp(name, next) == 0) {
            return env->objects[i];
        }
        next = env->strings[++i];
    }
    printf("DID NOT FIND SYMBOL\n");
    Object o;
    o.type = TYPE_ERROR;
    return o;
}

Result parse(struct Slice *slices)
{
    struct Slice *token = slices;
    struct Slice *rest;
    if(token->text != NULL) {
        rest = &slices[1];
    } else {
        printf("Assigning null...\n");
        rest = NULL;
    }

    if(token->text[0] == '(') {
        // todo check for null rest
        return readSeq(rest);
    } else { // todo error on closed paren
        return R(parseAtom(token), rest);
    }
}

Result readSeq(struct Slice *tokens)
{
    Object res = listObject();
    for(;;) {
        struct Slice *next = &tokens[0];
        struct Slice *rest = next->text? &next[1] : NULL;
        if(next->text[0] == ')') {
            return R(res, rest);
        }
        Result r = parse(tokens);
        addToList(&res, r.obj);
        tokens = r.slices;
    }
}

Object parseAtom(struct Slice *s)
{
    Object o;
    o.forward = NULL;
    if(isDigit(s->text[0])) {
        o.type = TYPE_NUMBER;
        o.number = 0;
        for(int i = 0; i < s->length; i++) {
            o.number *= 10;
            o.number += s->text[i] - '0';
        }
    } else if (s->text[0] == 'T' && s->text[1] == '\0') {
        o.type = TYPE_BOOL;
        o.number = 1;
    } else if (s->text[0] == 'F' && s->text[1] == '\0') {
        o.type = TYPE_BOOL;
        o.number = 0;
    } else {
        o.type = TYPE_SYMBOL;
        copySlice(o.name, s);
    }
    return o;
}

Object evalDefArgs(const Object *arg_forms, struct Environment *env)
{
    printf("evalDefArgs()\n");

    const Object *first_form = &arg_forms[0];
    const char *name = first_form->name;

    Object second_eval = eval(first_form->forward, env);

    addToEnv(env, name, second_eval);

    return *first_form;
}

Object evalIfArgs(const Object *arg_forms, struct Environment *env)
{
    return eval(arg_forms, env).number? 
            eval(arg_forms->forward, env) : 
            eval(arg_forms->forward->forward, env);
}

Object evalBuiltIns(const Object *first, const Object *rest, int *found,
                    struct Environment *env)
{
    if(strcmp(first->name, "def") == 0) {
        *found = 0;
        return evalDefArgs(rest, env);
    }else if(strcmp(first->name, "if") == 0) {
        *found = 0;
        return evalIfArgs(rest, env);
    }

    Object o;
    *found = -1;
    return o;
}

Object eval(const Object *obj, struct Environment *env)
{
    // printf("eval(): ");
    // printObj(obj);
    Object o = *obj;
    switch(obj->type) {
        case TYPE_NUMBER:
        case TYPE_BOOL:
            return o;

        case TYPE_SYMBOL:
            o = fetchFromEnvironment(obj->name, env);
            printf("fetched object '%s':\n", obj->name);
            printObj(&o);
            return o;

        case TYPE_LIST:
        {
            if(listLength(obj) == 1) {
                o = *obj->list;
                return o;
            }
            Object first_form = *obj->list;

            { // Try to eval built-ins
                int i = -1;
                Object built_in = 
                        evalBuiltIns(&first_form, first_form.forward, &i, env);

                if(i == 0)
                    return built_in;
            }
            printf("\nNo built-ins found\n");

            Object first_eval = eval(&first_form, env);
            Object arg1 = eval(first_form.forward, env);
            Object arg2 = eval(first_form.forward->forward, env);

            printf("Evaluating func\n");
            Object func_eval = first_eval.func(arg1, arg2);
            deleteList(obj);
            return func_eval;
        }
        default:
        ;
    }
    return o;
}

Result resultFromObjAndSlices(Object obj, struct Slice *slices)
{
    Result r;
    r.obj = obj;
    r.slices = slices;
    return r;
}

// todo could include and return a starting index for faster multi-adds
void addToEnv(struct Environment *env, const char *name, const Object obj)
{
    int i;
    for(i = 0; i < MAX_ENV_ELM; i++) {
        if(env->strings[i] == NULL) {
            env->strings[i] = malloc(sizeof(name));
            strncpy(env->strings[i], name, MAX_TOK_LEN);
            env->objects[i] = obj;
            break;
        }
    }
}

Object basicOp(Object *obj1, Object *obj2, const char op)
{
    Object o;
    o.forward = NULL;

    o.type = TYPE_NUMBER;
    const int n1 = obj1->number;
    const int n2 = obj2->number;

    switch(op){
        case '+':
            o.number = n1 + n2;
            return o;
        case '-':
            o.number = n1 - n2;
            return o;
        case '*':
            o.number = n1 * n2;
            return o;
        case '/':
            o.number = n1 / n2;
            return o;
    }

    o.type = TYPE_BOOL;
    switch(op) {
        case '=':
            o.number = n1 == n2;
            return o;
        case '>':
            o.number = n1 > n2;
            return o;
        case '<':
            o.number = n1 < n2;
            return o;
    }

    o = *obj1;
    return o;
}

#define bopf(_name, _char) \
Object _name(Object obj1, Object obj2) \
{ \
    return basicOp(&obj1, &obj2, _char); \
}

bopf(add, '+');
bopf(sub, '-');
bopf(mul, '*');
bopf(dvi, '/');
bopf(equ, '=');
bopf(gth, '>');
bopf(lth, '<');

void addFunc(const char *name, Object (*func)(Object, Object), 
               struct Environment *env)
{
    Object o;
    o.type = TYPE_FUNC;
    o.forward = NULL;
    o.func = func;
    addToEnv(env, name, o);
}

void deleteEnv(struct Environment *e)
{
    int i = 0;
    while(e->strings[i]) {
        free(e->strings[i]);
        i++;
    }
    free(e->strings);
    e->strings = NULL;

    free(e->objects);
    e->objects = NULL;
}

struct Environment defaultEnv() {
    struct Environment e;
    e.strings = malloc(sizeof(char*) * MAX_ENV_ELM);
    for(int i = 0; i < MAX_ENV_ELM; i++) {
        e.strings[i] = NULL;
    }

    e.objects = malloc(sizeof(Object) * MAX_ENV_ELM);

    addFunc("+", &add, &e);
    addFunc("-", &sub, &e);
    addFunc("*", &mul, &e);
    addFunc("/", &dvi, &e);
    addFunc("=", &equ, &e);
    addFunc(">", &gth, &e);
    addFunc("<", &lth, &e);

    return e;
}

Object parseEval(const char *input, struct Environment *env)
{
    struct Slice *tokens = nf_tokenize(input);
#ifdef STANDALONE
    struct Slice *debug = tokens;
    if(debug) {
        while(debug->text) {
            char tok[MAX_TOK_LEN];
            copySlice(tok, debug);
            debug++;
        }
    }
#endif
    Object parsed = parse(tokens).obj;
    free(tokens);
    return eval(&parsed, env);
}

#ifdef STANDALONE
int repl(struct Environment *env)
{
    char input[100] = "";
    while(input[0] != 'q') {
        printf("pebblisp>> ");
        fgets(input, 100, stdin);
        Object obj = parseEval(input, env);
        printObj(&obj);
        //break;
    }
}

int main(void)
{
    struct Environment env = defaultEnv();
#ifndef NO_REPL
    repl(&env);
#else
    struct Slice *tokens = nf_tokenize("(+ 10 5)");
    struct Slice *debug = tokens;

    if(debug) {
        while(debug->text) {
            char tok[10];
            copySlice(tok, debug);
            printf("'%s', ", tok);
            debug++;
        }
        printf("\n");
    } else {
        printf("parse error\n");
    }
    parse(tokens);
    free(tokens);
#endif
    deleteEnv(&env);
}
#endif
